/*
 *            DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
 *                    Version 2, December 2004
 *
 * Copyright (C) 2013 Willy "z3bra" Goiffon <willy@mailoo.org>
 *
 * Everyone is permitted to copy and distribute verbatim or modified
 * copies of this license document, and changing it is allowed as long
 * as the name is changed.
 *
 *            DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
 *   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION
 *
 *  0. You just DO WHAT THE FUCK YOU WANT TO.
 */

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/inotify.h>

#define BUF_LEN 512 /* buffer size for input */
#define DIR_LEN 255 /* directory lengh in char */
#define EVT_LEN (128 * (EVENT_SIZE + 16))
#define EVENT_SIZE (sizeof (struct inotify_event))
#define CHANS_LEN  (sizeof(channels)/sizeof(chan_t*))

#define ROOT_ALIAS "server"

/*
 * Represent a ii channel
 *  - index     : a unique index that represent the channel
 *  - alias     : a name given to the channel
 *  - input     : the path to its 'in' file
 */
typedef struct channel {
    int     index;
    char    alias[32];
    char    input[DIR_LEN];
    struct channel *next;
    struct channel *prev;
} chan_t;

char g_prompt = ']';    /* character used for prompt */
chan_t *g_head = NULL;

void usage()
{
    fputs(
            "usage: ptii [-h] [-i <irc_dir>] [-p <char>]\n"
            "\t-h            : display this help\n"
            "\t-i <irc_dir>  : use <irc_dir> instead of current directory\n"
            "\t-p <char>     : use <char> as the prompt\n"
            ,stdout);

    exit (0);
}

chan_t *chan_create(chan_t **head, char *root, char *alias)
{
    static int index = 0;
    chan_t  *c = NULL, *tmp = NULL;

    if (root) {
        /* allocate memory for a channel structure */
        c = calloc(1, sizeof(chan_t));

        /* remove the '/' at the end of the path */
        if (root[strnlen(root,DIR_LEN)-1] == '/') {
            root[strnlen(root,DIR_LEN)-1] = '\0';
        }
    }

    if (c) {
        /* Get the index */
        c->index = ++index;
        c->next = NULL;
        c->prev = NULL;

        /* Are we adding a channel, or the root? */
        if (alias && strnlen(alias, 32)) {
            strncpy(c->alias, alias, 32);
            snprintf(c->input, DIR_LEN, "%s/%s/in", root, alias);
        } else {
            strncpy(c->alias, ROOT_ALIAS, 32);
            snprintf(c->input, DIR_LEN, "%s/in", root);
        }

        /* Add "c" after all the struct */
        if (!(*head)) {
            (*head) = c;
        } else {
            for(tmp = (*head); tmp->next; tmp=tmp->next);
            c->prev=tmp;
            tmp->next = c;
            (*head)->prev = c;
        }


        return c;
    }

    return NULL;
}

void chan_add_exising(chan_t **head, char *root)
{
    DIR* rep = NULL;
    struct dirent *stdir = NULL;

    rep = opendir(root);
    if(rep != NULL) {

        while( (stdir = readdir(rep)) != NULL){
            if(stdir->d_type == DT_DIR &&
                    strcmp(stdir->d_name, ".") &&
                    strcmp(stdir->d_name, "..")){
                chan_create(head, root, stdir->d_name);
            }
        }

        if(closedir(rep) == -1){
            exit (-1);
        }
    } else {
        switch (errno) {
            case EACCES:
                fprintf(stderr,"bad permissions : %s/\n", root);
                break;
            case ENOTDIR:
                fprintf(stderr, "not a directory : %s.\n", root);
                break;
            default:
                fprintf(stderr, "erreur #%d\n",errno);
        }
    }
    return; /* void */
}

void chan_list(chan_t *node)
{
    if (!node) { return; }

    /* recurse to parse the whole list */
    printf("%d:%s (%s)\n", node->index, node->alias, node->input);
    chan_list (node->next);

    return; /* void */
}

void chan_free(chan_t **node)
{
    if ((*node)) {
        chan_free(&((*node)->next));
    }

    free((*node));

    return; /* void */
}

chan_t *chan_search_alias(chan_t *head, char *alias)
{
    if (head) {
        if (strncmp(head->alias, alias, 32)) {
            return head;
        }
        return chan_search_alias(head->next, alias);
    }
    return NULL;
}

chan_t *chan_search_index(chan_t *head, int index)
{
    if (head) {
        if (head->index == index) {
            return head;
        }
        return chan_search_index(head->next, index);
    }
    return NULL;
}

chan_t *chan_use_next(int *fd, chan_t *cur)
{
    int     new_fd = 0;

    if (cur->next) {
        cur = cur->next;
    } else if (g_head) {
        cur = g_head;
    } else {
        return cur;
    }

    new_fd = open(cur->input, O_WRONLY);

    if (new_fd > 0) {
        close(*fd);
        (*fd)  = new_fd;
    }

    return cur;
}

chan_t *chan_use_prev(int *fd, chan_t *cur)
{
    int     new_fd = 0;

    if (cur->prev) {
        cur = cur->prev;
    } else if (g_head) {
        cur = g_head;
    } else {
        return cur;
    }

    printf("switching to %s\n", cur->alias);
    new_fd = open(cur->input, O_WRONLY);

    if (new_fd > 0) {
        close(*fd);
        (*fd)  = new_fd;
    }

    return cur;
}

void inotify_event_parser(char *rootdir, int fd)
{
    int len, i=0;
    char *events = NULL;
    struct inotify_event *ev;

    /* enough size for 128 events */
    events = calloc(1, EVT_LEN);

    /* get how many bytes were read */
    len = read (fd, events,  EVT_LEN);

    if (len <= 0 ) {
        perror("read");
    }

    while (i < len) {

        /* get the event queued number i */
        ev = (struct inotify_event *) &events[i];

        /* Let's just handle directories here */
        if (ev->mask & IN_ISDIR && ev->len) {
            /* ii has just created a directory */
            if (ev->mask & IN_CREATE) {
                /* add the correspondig channel to memory */
                chan_create(&g_head, rootdir, ev->name);
            }
        }

        /* go on to next event */
        i += EVENT_SIZE + ev->len;
    }

    return; /* void */
}

void *inotify_handler(void *args)
{
    int fd = 0;
    int wd = 0;
    char *rootdir = NULL;

    /* cast args to be a char* pointed to by rootdir */
    rootdir = (char*) args;

    /* get a file descriptor to the watched directory */
    fd = inotify_init();
    if (fd < 0) { perror("inotify_init"); }

    /* watch for IN_CREATE events */
    wd = inotify_add_watch (fd, rootdir, IN_CREATE);
    if (wd < 0) { perror("inotify_add_watch"); exit(errno); }

    for (;;) {
        /* infinite loop to parse events */
        inotify_event_parser(rootdir, fd);
    }

    return NULL;
}

void run()
{
    int fd = 0;
    int write_to_pipe = 0;
    char buf[BUF_LEN];
    chan_t *cur = NULL;

    cur = chan_search_index(g_head, 1);
    fd = open(cur->input, O_WRONLY);

    while (!feof(stdin) && fd>0) {
        printf("%s%c ", cur->alias, g_prompt);
        fflush(stdout);
        fgets(buf, 255, stdin);

        write_to_pipe = 0;

        if (buf[0] == '/') {
            switch (buf[1]) {
                case 'n':
                    cur = chan_use_next(&fd, cur);
                    break;
                case 'p':
                    cur = chan_use_prev(&fd, cur);
                    break;
                case 'c':
                    /* parse channel-related commands */
                    switch (buf[2]) {
                        case 'l':
                            chan_list(g_head);
                            break;
                        default:
                            system("clear");
                    }
                    break;

                case 'x':
                    close(fd);
                    fd = 0;
                    break;

                default:
                    write_to_pipe = 1;
            }
        } else {
            write_to_pipe = 1;
        }

        if (write_to_pipe && fd>0 && write(fd, buf, strlen(buf)) < 0) {
            perror("write");
            exit(errno);
        }
    }

    return; /* void */
}

int main(int argc, char **argv)
{
    int i;
    pthread_t th;
    char cwd[DIR_LEN];

    /* put the current directory into cwd */
    if (getcwd(cwd, DIR_LEN) == NULL) {
        perror("getcwd");
        exit(errno);
    }

    /* is there is only the -h flag, exec usage() */
    if ((argc == 2 && argv[1][0] == '-' && argv[1][1] == 'h')) usage();

    /* parse the command like arguments */
    for(i = 1; (i + 1 < argc) && (argv[i][0] == '-'); i++) {
        switch (argv[i][1]) {
            /* put the directory in cwd */
            case 'i': strncpy(cwd, argv[++i], DIR_LEN); break;
            /* change the prompt */
            case 'p': g_prompt = argv[++i][0]; break;
        }
    }

    /* start a new thread that will watch the directory */
    pthread_create (&th, NULL, inotify_handler, (void*)cwd);

    /* start by creating a channel that represent the current dir */
    chan_create(&g_head, cwd, NULL);

    /* add a channel for each existing directory */
    chan_add_exising(&g_head, cwd);

    run();

    /* wait for the thread to terminate or force quitting */
    pthread_cancel(th);

    /* free the whole channel array */
    chan_free(&g_head);

    return EXIT_SUCCESS;
}
